@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2020 Red Hat Inc.
+ * Copyright Red Hat
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -136,7 +136,7 @@ check_perl_failure (void)
   return 0;
 }
 
-XS(xs_debug)
+XS (xs_debug)
 {
   dXSARGS;
   if (items >= 1)
@@ -146,7 +146,7 @@ XS(xs_debug)
 
 static int last_error;
 
-XS(xs_set_error)
+XS (xs_set_error)
 {
   dXSARGS;
   /* Is it worth adding error checking for bad arguments? */
@@ -171,28 +171,28 @@ xs_init (pTHX)
 
   /* This macro defines flags such as $Nbdkit::FLAG_MAY_TRIM
    * See also
-   * https://github.com/Perl/perl5/blob/5de22a40933c620b62e1f431457593777b99793d/os2/os2.c#L4563
+   * https://github.com/Perl/perl5/blob/5de22a40933c/os2/os2.c#L4563
    */
 #define DEFINE_FLAG(n)                           \
   gv = gv_fetchpv ("Nbdkit::" #n, TRUE, SVt_PV); \
   GvMULTI_on (gv);                               \
-  sv_setiv (GvSV(gv), NBDKIT_ ## n)
+  sv_setiv (GvSV (gv), NBDKIT_ ## n)
 
-  DEFINE_FLAG(FLAG_MAY_TRIM);
-  DEFINE_FLAG(FLAG_FUA);
-  DEFINE_FLAG(FLAG_REQ_ONE);
-  DEFINE_FLAG(FLAG_FAST_ZERO);
+  DEFINE_FLAG (FLAG_MAY_TRIM);
+  DEFINE_FLAG (FLAG_FUA);
+  DEFINE_FLAG (FLAG_REQ_ONE);
+  DEFINE_FLAG (FLAG_FAST_ZERO);
 
-  DEFINE_FLAG(FUA_NONE);
-  DEFINE_FLAG(FUA_EMULATE);
-  DEFINE_FLAG(FUA_NATIVE);
+  DEFINE_FLAG (FUA_NONE);
+  DEFINE_FLAG (FUA_EMULATE);
+  DEFINE_FLAG (FUA_NATIVE);
 
-  DEFINE_FLAG(CACHE_NONE);
-  DEFINE_FLAG(CACHE_EMULATE);
-  DEFINE_FLAG(CACHE_NATIVE);
+  DEFINE_FLAG (CACHE_NONE);
+  DEFINE_FLAG (CACHE_EMULATE);
+  DEFINE_FLAG (CACHE_NATIVE);
 
-  DEFINE_FLAG(EXTENT_HOLE);
-  DEFINE_FLAG(EXTENT_ZERO);
+  DEFINE_FLAG (EXTENT_HOLE);
+  DEFINE_FLAG (EXTENT_ZERO);
 }
 
 static void
@@ -339,6 +339,21 @@ perl_open (int readonly)
   SV *sv;
   dSP;
 
+  /*
+   * From perlembed(1):
+   *
+   * "PERL_SET_CONTEXT(interp) should also be called whenever "interp"
+   * is used by a thread that did not create it (using either
+   * perl_alloc(), or the more esoteric perl_clone())."
+   *
+   * Since we may be called here from a new thread created within
+   * nbdkit, do this.  This is necessary since Perl 5.38.  It didn't
+   * seem to make a difference with earlier Perl, but doesn't break
+   * them either.
+   */
+
+  PERL_SET_CONTEXT (my_perl);
+
   /* We check in perl_config that this callback is defined. */
   ENTER;
   SAVETMPS;
@@ -685,4 +700,4 @@ static struct nbdkit_plugin plugin = {
   .zero              = perl_zero,
 };
 
-NBDKIT_REGISTER_PLUGIN(plugin)
+NBDKIT_REGISTER_PLUGIN (plugin)
