@@ -52,8 +52,8 @@
 //!         "my_plugin"
 //!     }
 //!
-//!     fn open(_readonly: bool) -> Box<dyn Server> {
-//!         Box::new(MyPlugin::default())
+//!     fn open(_readonly: bool) -> Result<Box<dyn Server>> {
+//!         Ok(Box::new(MyPlugin::default()))
 //!     }
 //!
 //!     fn read_at(&self, buf: &mut [u8], offset: u64) -> Result<()> {
@@ -76,7 +76,7 @@ use std::{
     fmt,
     io,
     mem,
-    os::raw::{c_char, c_int, c_void},
+    os::raw::{c_char, c_int, c_double, c_void},
     ptr,
     slice,
     sync::Once,
@@ -138,11 +138,12 @@ mod unreachable {
     pub(super) fn config(_k: &str, _v: &str) -> Result<()> { unreachable!() }
     pub(super) fn config_complete() -> Result<()> { unreachable!() }
     pub(super) fn dump_plugin() { unreachable!() }
-    pub(super) fn open(_: bool) -> Box<dyn Server> { unreachable!() }
+    pub(super) fn open(_: bool) -> Result<Box<dyn Server>> { unreachable!() }
     pub(super) fn preconnect(_: bool) -> Result<()> { unreachable!() }
     pub(super) fn thread_model() -> Result<ThreadModel> { unreachable!() }
 }
 
+static mut AFTER_FORK: fn() -> Result<()> = unreachable::config_complete;
 static mut CONFIG: fn(k: &str, v: &str) -> Result<()> = unreachable::config;
 static mut CONFIG_COMPLETE: fn() -> Result<()> = unreachable::config_complete;
 static mut CONFIG_HELP: Vec<u8> = Vec::new();
@@ -153,7 +154,7 @@ static mut LOAD: fn() = unreachable::dump_plugin;
 static mut LONGNAME: Vec<u8> = Vec::new();
 static mut MAGIC_CONFIG_KEY: Vec<u8> = Vec::new();
 static mut NAME: Vec<u8> = Vec::new();
-static mut OPEN: fn(readonly: bool) -> Box<dyn Server> = unreachable::open;
+static mut OPEN: fn(readonly: bool) -> Result<Box<dyn Server>> = unreachable::open;
 static mut PRECONNECT: fn(readonly: bool) -> Result<()> = unreachable::preconnect;
 static mut THREAD_MODEL: fn() -> Result<ThreadModel> = unreachable::thread_model;
 static mut UNLOAD: fn() = unreachable::dump_plugin;
@@ -315,6 +316,35 @@ impl ExtentHandle {
 mod ffi {
     use super::*;
 
+    pub(super) extern fn after_fork() -> c_int {
+        match unsafe { AFTER_FORK() } {
+            Ok(()) => 0,
+            Err(e) => {
+                set_error(e);
+                -1
+            }
+        }
+    }
+
+    pub(super) extern fn block_size(h: *mut c_void,
+                                    minp: *mut u32,
+                                    prefp: *mut u32,
+                                    maxp: *mut u32) -> c_int {
+        let server = unsafe { downcast(h) };
+        match server.block_size() {
+            Err(e) => {
+                set_error(e);
+                -1
+            },
+            Ok((min, pref, max)) => unsafe {
+                *minp = min;
+                *prefp = pref;
+                *maxp = max;
+                0
+            }
+        }
+    }
+
     macro_rules! can_method {
         ( $meth:ident ) => {
             pub(super) extern fn $meth(h: *mut c_void) -> c_int {
@@ -493,9 +523,15 @@ mod ffi {
     pub(super) extern fn open(readonly: c_int) -> *mut c_void {
         // We need to double-box to turn the trait object (fat pointer) into a
         // thin pointer
-        let server = Box::new(unsafe{OPEN(readonly != 0)});
-        // Leak the memory to C.  We'll get it back in close
-        Box::into_raw(server) as *mut c_void
+        match unsafe { OPEN(readonly != 0) } {
+            Ok(server) =>
+                // Leak the memory to C.  We'll get it back in close
+                Box::into_raw(Box::new(server)) as *mut c_void,
+            Err(e) => {
+                set_error(e);
+                ptr::null_mut()
+            }
+        }
     }
 
     pub(super) extern fn pread(h: *mut c_void,
@@ -584,6 +620,25 @@ mod ffi {
 // We want the argument names to show up without underscores in the API docs
 #[allow(unused_variables)]
 pub trait Server {
+    /// This optional callback is called before the server starts serving.
+    ///
+    /// It is called after the server forks and changes directory. If
+    /// a plugin needs to create background threads it should do so here.
+    fn after_fork() -> Result<()> where Self: Sized { unimplemented!() }
+
+    /// Return block size hints to the client.
+    ///
+    /// This is called during option negotiation to get the
+    /// minimum, preferred and maximum block size (all in bytes).
+    /// The client should (but is not required to) obey these
+    /// constraints by not making requests smaller than the
+    /// minimum block size or larger than the maximum size
+    /// and usually making requests of a multiple of the
+    /// preferred size.  Additionally requests should be
+    /// aligned to the minimum (and usually the preferred)
+    /// block size.
+    fn block_size(&self) -> Result<(u32, u32, u32)> { unimplemented!() }
+
     /// Indicates that the client intends to make further accesses to the given
     /// data region.
     ///
@@ -801,7 +856,7 @@ pub trait Server {
     /// this flag is false) by returning false from the [`Server::can_write`]
     /// callback.  So if your plugin can only serve read-only, you can ignore
     /// this parameter.
-    fn open(readonly: bool) -> Box<dyn Server> where Self: Sized;
+    fn open(readonly: bool) -> Result<Box<dyn Server>> where Self: Sized;
 
     /// This optional callback is called when a TCP connection has been made to
     /// the server. This happens early, before NBD or TLS negotiation. If TLS
@@ -900,6 +955,8 @@ macro_rules! opt_method {
 #[doc(hidden)]
 #[derive(Default)]
 pub struct Builder {
+    pub after_fork: bool,
+    pub block_size: bool,
     pub cache: bool,
     pub can_cache: bool,
     pub can_extents: bool,
@@ -938,6 +995,7 @@ impl Builder {
                 CONFIG_COMPLETE = S::config_complete;
                 DUMP_PLUGIN = S::dump_plugin;
                 GET_READY = S::get_ready;
+                AFTER_FORK = S::after_fork;
                 LOAD = S::load;
                 OPEN = S::open;
                 PRECONNECT = S::preconnect;
@@ -1021,7 +1079,13 @@ impl Builder {
             thread_model: opt_method!(self, thread_model),
             can_fast_zero: opt_method!(self, can_fast_zero),
             preconnect: opt_method!(self, preconnect),
-            get_ready: opt_method!(self, get_ready)
+            get_ready: opt_method!(self, get_ready),
+            after_fork: opt_method!(self, after_fork),
+            _list_exports: None, // not implemented
+            _default_export: ptr::null(), // not implemented
+            _export_description: ptr::null(), // not implemented
+            _cleanup: None, // not implemented
+            block_size: opt_method!(self, block_size)
         };
         // Leak the memory to C.  NBDKit will never give it back.
         Box::into_raw(Box::new(plugin))
@@ -1039,6 +1103,8 @@ impl Builder {
 extern "C" {
     fn nbdkit_add_extent(extents: *mut c_void, offset: u64, length: u64,
                          ty: u32) -> c_int;
+    #[doc(hidden)]
+    pub fn nbdkit_debug(fmt: *const c_char, ...);
     fn nbdkit_error(fmt: *const c_char, ...);
     fn nbdkit_export_name() -> *const c_char;
     fn nbdkit_set_error(errno: c_int);
@@ -1048,6 +1114,35 @@ extern "C" {
     fn nbdkit_shutdown();
     fn nbdkit_disconnect(force: bool);
     fn nbdkit_stdio_safe() -> c_int;
+    fn nbdkit_is_tls() -> c_int;
+    fn nbdkit_parse_bool(s: *const c_char) -> c_int;
+    fn nbdkit_parse_size(s: *const c_char) -> i64;
+    fn nbdkit_parse_probability(what: *const c_char,
+                                s: *const c_char,
+                                retp: *mut c_double) -> c_int;
+}
+
+/// Print a debug message.
+///
+/// Debug messages are only output when the server runs in verbose mode
+/// (-v flag).  They are sent to stderr if nbdkit is running in the
+/// foreground, or to syslog if nbdkit is forked into the background.
+#[macro_export]
+macro_rules! debug {
+    ($($args:tt)*) => {
+        match ::std::ffi::CString::new(format!($($args)*)) {
+            Ok(msg) =>
+                unsafe {
+                    $crate::nbdkit_debug(b"%s\0".as_ptr()
+                                             as *const ::libc::c_char,
+                                         msg.as_ptr());
+                },
+            Err(_) =>
+                // probably the string contained \0, ignore as it
+                // is only a debug message
+                ()
+        }
+    }
 }
 
 /// Return the optional NBD export name if one was negotiated with the current
@@ -1074,7 +1169,24 @@ pub fn export_name() -> std::result::Result<String, Box<dyn error::Error>> {
 /// configuration phase, the client should assume that stdin and stdout have
 /// been closed.
 pub fn is_stdio_safe() -> bool {
-    unsafe { nbdkit_stdio_safe() == 1 }
+    unsafe { nbdkit_stdio_safe() != 0 } // note: cannot return an error
+}
+
+/// Did the client complete TLS authentication?
+///
+/// Call this from the `open` method.  It returns true if the client
+/// completed TLS authentication, or false if the connection is using
+/// plaintext.
+pub fn is_tls() -> Result<bool> {
+    let r = unsafe { nbdkit_is_tls() };
+    match r {
+        0 => Ok(false),
+        x if x > 0 => Ok(true),
+        _ =>
+            // nbdkit_error was called, so the real error was already
+            // written on stderr or sent to the logs.
+            Err(Error::new(libc::EINVAL, "nbdkit_is_tls failed"))
+    }
 }
 
 /// Return the peer (client) address, if available.
@@ -1112,6 +1224,60 @@ pub fn disconnect(force: bool) {
     unsafe { nbdkit_disconnect(force) };
 }
 
+/// Parse a string as a boolean using `nbdkit_parse_bool`.
+///
+/// This parses boolean strings like `true`, `no`, etc in the same way as
+/// other (non Rust) nbdkit plugins, so you can use this function for
+/// consistency with other plugins.
+pub fn parse_bool(s: &str) -> Result<bool> {
+    let cs = CString::new(s).unwrap();
+    let r = unsafe { nbdkit_parse_bool(cs.as_ptr()) };
+    match r {
+        -1 =>
+            // nbdkit_error was called already.
+            Err(Error::new(libc::EINVAL,
+                           "failed to parse boolean using nbdkit_parse_bool")),
+        0 => Ok(false),
+        _ => Ok(true)
+    }
+}
+
+/// Parse a string as a size using `nbdkit_parse_size`.
+///
+/// This parses size strings like `1M`, `7E`, etc in the same way as
+/// other (non Rust) nbdkit plugins, so you can use this function for
+/// consistency with other plugins.
+pub fn parse_size(s: &str) -> Result<i64> {
+    let cs = CString::new(s).unwrap();
+    let r = unsafe { nbdkit_parse_size(cs.as_ptr()) };
+    if r == -1 {
+        // nbdkit_error was called already.
+        return Err(
+            Error::new(libc::EINVAL,
+                       "failed to parse size using nbdkit_parse_size"));
+    }
+    Ok(r)
+}
+
+/// Parse a string as a probability/percentage using
+/// `nbdkit_parse_probability`.
+///
+/// This parses probabilities like `0.1`, `1e-1`, `10%`, `1:10`, `1/10`.
+pub fn parse_probability(what: &str, s: &str) -> Result<f64> {
+    let ws = CString::new(what).unwrap();
+    let cs = CString::new(s).unwrap();
+    let mut d = 0.0;
+    let r = unsafe { nbdkit_parse_probability(ws.as_ptr(), cs.as_ptr(),
+                                              &mut d) };
+    if r == -1 {
+        // nbdkit_error was called already.
+        return Err(
+            Error::new(libc::EINVAL,
+                       "failed to parse probability using nbdkit_parse_probability"));
+    }
+    Ok(d)
+}
+
 #[doc(hidden)]
 #[repr(C)]
 pub struct Plugin {
@@ -1191,6 +1357,19 @@ pub struct Plugin {
     pub can_fast_zero: Option<extern fn (h: *mut c_void) -> c_int>,
     pub preconnect: Option<extern fn(readonly: c_int) -> c_int>,
     pub get_ready: Option<extern fn() -> c_int>,
+    pub after_fork: Option<extern fn() -> c_int>,
+
+    // These fields exist in the struct but have not been implemented
+    // by the Rust bindings yet.
+    #[doc(hidden)] pub _list_exports: Option<extern fn ()>,
+    #[doc(hidden)] pub _default_export: *const c_char,
+    #[doc(hidden)] pub _export_description: *const c_char,
+    #[doc(hidden)] pub _cleanup: Option<extern fn ()>,
+
+    pub block_size: Option<extern fn(h: *mut c_void,
+                                     minp: *mut u32,
+                                     prefp: *mut u32,
+                                     maxp: *mut u32) -> c_int>,
 }
 
 /// Register your plugin with NBDKit.
@@ -1214,7 +1393,7 @@ pub struct Plugin {
 ///         "my_plugin"
 ///     }
 ///
-///     fn open(_readonly: bool) -> Box<dyn Server> {
+///     fn open(_readonly: bool) -> Result<Box<dyn Server>> {
 ///         # unimplemented!();
 ///         // ...
 ///     }
@@ -1301,9 +1480,9 @@ mod t {
                     unsafe {
                         *sin = libc::sockaddr_in {
                             sin_family: libc::AF_INET as libc::sa_family_t,
-                            sin_port: u16::from_le_bytes([4, 0xd2]),
+                            sin_port: u16::from_ne_bytes([4, 0xd2]),
                             sin_addr: libc::in_addr {
-                                s_addr: u32::from_le_bytes([127, 0, 0, 1])
+                                s_addr: u32::from_ne_bytes([127, 0, 0, 1])
                             },
                             .. mem::zeroed()
                         };
@@ -1329,7 +1508,7 @@ mod t {
                     unsafe {
                         *sin6 = libc::sockaddr_in6 {
                             sin6_family: libc::AF_INET6 as libc::sa_family_t,
-                            sin6_port: u16::from_le_bytes([4, 0xd2]),
+                            sin6_port: u16::from_ne_bytes([4, 0xd2]),
                             sin6_addr: libc::in6_addr {
                                 s6_addr: [0, 0, 0, 0, 0, 0, 0, 0,
                                          0, 0, 0, 0, 0, 0, 0, 1],
