@@ -66,6 +66,7 @@
 #include "ascii-string.h"
 #include "exit-with-parent.h"
 #include "nbd-protocol.h"
+#include "open_memstream.h"
 #include "realpath.h"
 #include "strndup.h"
 #include "syslog.h"
@@ -79,6 +80,7 @@
 #endif
 
 static char *make_random_fifo (void);
+static char *make_uri (void);
 static struct backend *open_plugin_so (size_t i, const char *filename,
                                        int short_name);
 static struct backend *open_filter_so (struct backend *next, size_t i,
@@ -117,6 +119,8 @@ const char *user, *group;       /* -u & -g */
 bool verbose;                   /* -v */
 bool vsock;                     /* --vsock */
 unsigned int socket_activation; /* $LISTEN_FDS and $LISTEN_PID set */
+enum service_mode service_mode; /* serving over TCP, Unix, etc */
+char *uri;                      /* NBD URI */
 bool configured;                /* .config_complete done */
 int saved_stdin = -1;           /* dup'd stdin during -s/--run */
 int saved_stdout = -1;          /* dup'd stdout during -s/--run */
@@ -186,6 +190,7 @@ dump_config (void)
   printf ("%s=%s\n", "name", PACKAGE_NAME);
   printf ("%s=%s\n", "plugindir", plugindir);
   printf ("%s=%s\n", "root_tls_certificates_dir", root_tls_certificates_dir);
+  printf ("%s=%s\n", "run_default_socket", "Unix");
   printf ("%s=%s\n", "sbindir", sbindir);
 #ifdef HAVE_LIBSELINUX
   printf ("selinux=yes\n");
@@ -210,6 +215,22 @@ dump_config (void)
 #endif
 }
 
+/* -D nbdkit.environ=1 to dump the environment at start up. */
+NBDKIT_DLL_PUBLIC int nbdkit_debug_environ;
+
+#ifndef HAVE_ENVIRON_DECL
+extern char **environ;
+#endif
+
+static void
+dump_environment (void)
+{
+  size_t i;
+
+  for (i = 0; environ[i]; ++i)
+    nbdkit_debug ("%s", environ[i]);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -601,6 +622,38 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
+  /* Since nbdkit 1.36, --run implies -U -, unless --vsock or --port
+   * was set explicitly.
+   */
+  if (run && !unixsocket && !port && !vsock) {
+    unixsocket = make_random_fifo ();
+    if (!unixsocket)
+      exit (EXIT_FAILURE);
+  }
+
+  /* By the point we have enough information to calculate the service mode. */
+  if (socket_activation)
+    service_mode = SERVICE_MODE_SOCKET_ACTIVATION;
+  else if (listen_stdin)
+    service_mode = SERVICE_MODE_LISTEN_STDIN;
+  else if (unixsocket)
+    service_mode = SERVICE_MODE_UNIXSOCKET;
+  else if (vsock)
+    service_mode = SERVICE_MODE_VSOCK;
+  else
+    service_mode = SERVICE_MODE_TCPIP;
+  debug ("service mode: %s", service_mode_string (service_mode));
+
+  /* By the point we have enough information to calculate the NBD URI.
+   * Note this may be NULL.
+   */
+  uri = make_uri ();
+  if (uri)
+    debug ("NBD URI: %s", uri);
+  else
+    debug ("no NBD URI because service mode is %s",
+           service_mode_string (service_mode));
+
   /* The remaining command line arguments are the plugin name and
    * parameters.  If --help, --version or --dump-plugin were specified
    * then we open the plugin so that we can display the per-plugin
@@ -662,6 +715,12 @@ main (int argc, char *argv[])
   /* Check all debug flags were used, and free them. */
   free_debug_flags ();
 
+  /* Dump the environment if asked.  This is the earliest we can do it
+   * because it uses a debug flag.
+   */
+  if (nbdkit_debug_environ && verbose)
+    dump_environment ();
+
   if (help) {
     struct backend *b;
 
@@ -764,6 +823,7 @@ main (int argc, char *argv[])
 
   free (unixsocket);
   free (pidfile);
+  free (uri);
 
   if (random_fifo) {
     unlink (random_fifo);
@@ -821,6 +881,90 @@ make_random_fifo (void)
   return NULL;
 }
 
+const char *
+service_mode_string (enum service_mode sm)
+{
+  switch (sm) {
+  case SERVICE_MODE_SOCKET_ACTIVATION: return "socket activation";
+  case SERVICE_MODE_LISTEN_STDIN:      return "listen stdin";
+  case SERVICE_MODE_UNIXSOCKET:        return "unix socket";
+  case SERVICE_MODE_VSOCK:             return "vsock";
+  case SERVICE_MODE_TCPIP:             return "TCP/IP";
+  default: abort ();
+  }
+}
+
+static char *
+make_uri (void)
+{
+  FILE *fp;
+  size_t len = 0;
+  char *r = NULL;
+
+  switch (service_mode) {
+  case SERVICE_MODE_SOCKET_ACTIVATION:
+  case SERVICE_MODE_LISTEN_STDIN:
+    /* can't form a URI, uri will be NULL */
+    return NULL;
+  default: ;
+  }
+
+  fp = open_memstream (&r, &len);
+  if (fp == NULL) {
+    perror ("open_memstream");
+    exit (EXIT_FAILURE);
+  }
+
+  switch (service_mode) {
+  case SERVICE_MODE_UNIXSOCKET:
+    fprintf (fp, "nbd%s+unix://", tls == 2 ? "s" : "");
+    if (export_name && strcmp (export_name, "") != 0) {
+      putc ('/', fp);
+      uri_quote (export_name, fp);
+    }
+    fprintf (fp, "?socket=");
+    uri_quote (unixsocket, fp);
+    break;
+  case SERVICE_MODE_VSOCK:
+    /* 1 = VMADDR_CID_LOCAL */
+    fprintf (fp, "nbd%s+vsock://1", tls == 2 ? "s" : "");
+    if (port) {
+      putc (':', fp);
+      fputs (port, fp);
+    }
+    if (export_name && strcmp (export_name, "") != 0) {
+      putc ('/', fp);
+      uri_quote (export_name, fp);
+    }
+    break;
+  case SERVICE_MODE_TCPIP:
+    fprintf (fp, "nbd%s://localhost", tls == 2 ? "s" : "");
+    if (port) {
+      putc (':', fp);
+      fputs (port, fp);
+    }
+    if (export_name && strcmp (export_name, "") != 0) {
+      putc ('/', fp);
+      uri_quote (export_name, fp);
+    }
+    break;
+
+  case SERVICE_MODE_SOCKET_ACTIVATION:
+  case SERVICE_MODE_LISTEN_STDIN:
+    /* these labels are not-reachable, see above */
+    /*FALLTHROUGH*/
+  default:
+    abort ();
+  }
+
+  if (fclose (fp) == EOF) {
+    perror ("memstream failed");
+    exit (EXIT_FAILURE);
+  }
+
+  return r;
+}
+
 static struct backend *
 open_plugin_so (size_t i, const char *name, int short_name)
 {
@@ -948,15 +1092,16 @@ start_serving (void)
 #endif
   }
 
-  /* Socket activation: the ‘socket_activation’ variable (> 0) is the
-   * number of file descriptors from FIRST_SOCKET_ACTIVATION_FD to
-   * FIRST_SOCKET_ACTIVATION_FD+socket_activation-1.
-   */
-  if (socket_activation) {
-      if (sockets_reserve (&socks, socket_activation) == -1) {
-        perror ("realloc");
-        exit (EXIT_FAILURE);
-      }
+  switch (service_mode) {
+  case SERVICE_MODE_SOCKET_ACTIVATION:
+    /* Socket activation: the ‘socket_activation’ variable (> 0) is
+     * the number of file descriptors from FIRST_SOCKET_ACTIVATION_FD
+     * to FIRST_SOCKET_ACTIVATION_FD+socket_activation-1.
+     */
+    if (sockets_reserve (&socks, socket_activation) == -1) {
+      perror ("realloc");
+      exit (EXIT_FAILURE);
+    }
     for (i = 0; i < socket_activation; ++i) {
       int s = FIRST_SOCKET_ACTIVATION_FD + i, r;
       /* This can't fail because of the reservation above. */
@@ -968,35 +1113,42 @@ start_serving (void)
     write_pidfile ();
     top->after_fork (top);
     accept_incoming_connections (&socks);
-    return;
-  }
+    break;
 
-  /* Handling a single connection on stdin/stdout. */
-  if (listen_stdin) {
+  case SERVICE_MODE_LISTEN_STDIN:
+    /* Handling a single connection on stdin/stdout. */
     change_user ();
     write_pidfile ();
     top->after_fork (top);
     threadlocal_new_server_thread ();
     handle_single_connection (saved_stdin, saved_stdout);
-    return;
-  }
+    break;
 
-  /* Handling multiple connections on TCP/IP, Unix domain socket or
-   * AF_VSOCK.
-   */
-  if (unixsocket)
+  case SERVICE_MODE_UNIXSOCKET:
     bind_unix_socket (&socks);
-  else if (vsock)
-    bind_vsock (&socks);
-  else
-    bind_tcpip_socket (&socks);
+    goto serve_socks;
 
-  run_command ();
-  change_user ();
-  fork_into_background ();
-  write_pidfile ();
-  top->after_fork (top);
-  accept_incoming_connections (&socks);
+  case SERVICE_MODE_VSOCK:
+    bind_vsock (&socks);
+    goto serve_socks;
+
+  case SERVICE_MODE_TCPIP:
+    bind_tcpip_socket (&socks);
+  serve_socks:
+    /* Common code for handling multiple connections on TCP/IP, Unix
+     * domain socket or AF_VSOCK.
+     */
+    run_command ();
+    change_user ();
+    fork_into_background ();
+    write_pidfile ();
+    top->after_fork (top);
+    accept_incoming_connections (&socks);
+    break;
+
+  default:
+    abort ();
+  }
 }
 
 static void
