@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2020 Red Hat Inc.
+ * Copyright Red Hat
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -46,6 +46,7 @@
 
 #include "internal.h"
 #include "minmax.h"
+#include "ispowerof2.h"
 
 /* We extend the generic backend struct with extra fields relating
  * to this plugin.
@@ -164,6 +165,7 @@ plugin_dump_fields (struct backend *b)
   HAS (thread_model);
   HAS (get_ready);
   HAS (after_fork);
+  HAS (cleanup);
   HAS (preconnect);
   HAS (list_exports);
   HAS (default_export);
@@ -172,6 +174,7 @@ plugin_dump_fields (struct backend *b)
   HAS (close);
   HAS (export_description);
   HAS (get_size);
+  HAS (block_size);
   HAS (can_write);
   HAS (can_flush);
   HAS (is_rotational);
@@ -273,6 +276,17 @@ plugin_after_fork (struct backend *b)
     exit (EXIT_FAILURE);
 }
 
+static void
+plugin_cleanup (struct backend *b)
+{
+  struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
+
+  debug ("%s: cleanup", b->name);
+
+  if (p->plugin.cleanup)
+    p->plugin.cleanup ();
+}
+
 static int
 plugin_preconnect (struct backend *b, int readonly)
 {
@@ -310,15 +324,14 @@ plugin_default_export (struct backend *b, int readonly, int is_tls)
 }
 
 static void *
-plugin_open (struct backend *b, int readonly, const char *exportname,
+plugin_open (struct context *c, int readonly, const char *exportname,
              int is_tls)
 {
-  GET_CONN;
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   void *r;
 
   assert (p->plugin.open != NULL);
-  assert (conn->exportname == NULL);
 
   /* Save the exportname since the lifetime of the string passed in
    * here is likely to be brief.  In addition this provides a place
@@ -332,13 +345,16 @@ plugin_open (struct backend *b, int readonly, const char *exportname,
    * will still need to save the export name in the handle because of
    * the lifetime issue.
    */
-  conn->exportname = nbdkit_strdup_intern (exportname);
-  if (conn->exportname == NULL)
-    return NULL;
+  if (c->conn) {
+    assert (c->conn->exportname == NULL);
+    c->conn->exportname = nbdkit_strdup_intern (exportname);
+    if (c->conn->exportname == NULL)
+      return NULL;
+  }
 
   r = p->plugin.open (readonly);
-  if (r == NULL)
-    conn->exportname = NULL;
+  if (r == NULL && c->conn)
+    c->conn->exportname = NULL;
   return r;
 }
 
@@ -347,48 +363,110 @@ plugin_open (struct backend *b, int readonly, const char *exportname,
  * .close.
  */
 static int
-plugin_prepare (struct backend *b, void *handle,
-                int readonly)
+plugin_prepare (struct context *c, int readonly)
 {
   return 0;
 }
 
 static int
-plugin_finalize (struct backend *b, void *handle)
+plugin_finalize (struct context *c)
 {
   return 0;
 }
 
 static void
-plugin_close (struct backend *b, void *handle)
+plugin_close (struct context *c)
 {
-  GET_CONN;
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
-  if (handle && p->plugin.close)
-    p->plugin.close (handle);
-  conn->exportname = NULL;
+  assert (c->handle);
+  if (p->plugin.close)
+    p->plugin.close (c->handle);
+  if (c->conn)
+    c->conn->exportname = NULL;
 }
 
 static const char *
-plugin_export_description (struct backend *b, void *handle)
+plugin_export_description (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.export_description)
-    return p->plugin.export_description (handle);
+    return p->plugin.export_description (c->handle);
   else
     return NULL;
 }
 
 static int64_t
-plugin_get_size (struct backend *b, void *handle)
+plugin_get_size (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   assert (p->plugin.get_size != NULL);
 
-  return p->plugin.get_size (handle);
+  return p->plugin.get_size (c->handle);
+}
+
+static int
+plugin_block_size (struct context *c,
+                   uint32_t *minimum, uint32_t *preferred, uint32_t *maximum)
+{
+  struct backend *b = c->b;
+  struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
+
+  if (p->plugin.block_size) {
+    if (p->plugin.block_size (c->handle, minimum, preferred, maximum) == -1)
+      return -1;
+
+    /* To make scripting easier, it's permitted to set
+     * minimum = preferred = maximum = 0 and return 0.
+     * That means "no information", and works the same
+     * way as the else clause below.
+     */
+    if (*minimum == 0 && *preferred == 0 && *maximum == 0)
+      return 0;
+
+    if (*minimum < 1 || *minimum > 65536) {
+      nbdkit_error ("plugin must set minimum block size between 1 and 64K");
+      return -1;
+    }
+    if (! is_power_of_2 (*minimum)) {
+      nbdkit_error ("plugin must set minimum block size to a power of 2");
+      return -1;
+    }
+    if (! is_power_of_2 (*preferred)) {
+      nbdkit_error ("plugin must set preferred block size to a power of 2");
+      return -1;
+    }
+    if (*preferred < 512 || *preferred > 32 * 1024 * 1024) {
+      nbdkit_error ("plugin must set preferred block size "
+                    "between 512 and 32M");
+      return -1;
+    }
+    if (*maximum != (uint32_t)-1 && (*maximum % *minimum) != 0) {
+      nbdkit_error ("plugin must set maximum block size "
+                    "to -1 or a multiple of minimum block size");
+      return -1;
+    }
+    if (*minimum > *preferred || *preferred > *maximum) {
+      nbdkit_error ("plugin must set minimum block size "
+                    "<= preferred <= maximum");
+      return -1;
+    }
+
+    return 0;
+  }
+  else {
+    /* If there is no .block_size call then return minimum = preferred
+     * = maximum = 0, which is a sentinel meaning don't send the
+     * NBD_INFO_BLOCK_SIZE message.
+     */
+    *minimum = *preferred = *maximum = 0;
+    return 0;
+  }
 }
 
 static int
@@ -401,52 +479,57 @@ normalize_bool (int value)
 }
 
 static int
-plugin_can_write (struct backend *b, void *handle)
+plugin_can_write (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_write)
-    return normalize_bool (p->plugin.can_write (handle));
+    return normalize_bool (p->plugin.can_write (c->handle));
   else
     return p->plugin.pwrite || p->plugin._pwrite_v1;
 }
 
 static int
-plugin_can_flush (struct backend *b, void *handle)
+plugin_can_flush (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_flush)
-    return normalize_bool (p->plugin.can_flush (handle));
+    return normalize_bool (p->plugin.can_flush (c->handle));
   else
     return p->plugin.flush || p->plugin._flush_v1;
 }
 
 static int
-plugin_is_rotational (struct backend *b, void *handle)
+plugin_is_rotational (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.is_rotational)
-    return normalize_bool (p->plugin.is_rotational (handle));
+    return normalize_bool (p->plugin.is_rotational (c->handle));
   else
     return 0; /* assume false */
 }
 
 static int
-plugin_can_trim (struct backend *b, void *handle)
+plugin_can_trim (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_trim)
-    return normalize_bool (p->plugin.can_trim (handle));
+    return normalize_bool (p->plugin.can_trim (c->handle));
   else
     return p->plugin.trim || p->plugin._trim_v1;
 }
 
 static int
-plugin_can_zero (struct backend *b, void *handle)
+plugin_can_zero (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
@@ -455,7 +538,7 @@ plugin_can_zero (struct backend *b, void *handle)
    * expects .can_zero to return a tri-state on level of support.
    */
   if (p->plugin.can_zero) {
-    r = p->plugin.can_zero (handle);
+    r = p->plugin.can_zero (c->handle);
     if (r == -1)
       return -1;
     return r ? NBDKIT_ZERO_NATIVE : NBDKIT_ZERO_EMULATE;
@@ -466,46 +549,49 @@ plugin_can_zero (struct backend *b, void *handle)
 }
 
 static int
-plugin_can_fast_zero (struct backend *b, void *handle)
+plugin_can_fast_zero (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
   if (p->plugin.can_fast_zero)
-    return normalize_bool (p->plugin.can_fast_zero (handle));
+    return normalize_bool (p->plugin.can_fast_zero (c->handle));
   /* Advertise support for fast zeroes if no .zero or .can_zero is
    * false: in those cases, we fail fast instead of using .pwrite.
    * This also works when v1 plugin has only ._zero_v1.
    */
   if (p->plugin.zero == NULL)
     return 1;
-  r = backend_can_zero (b);
+  r = backend_can_zero (c);
   if (r == -1)
     return -1;
   return !r;
 }
 
 static int
-plugin_can_extents (struct backend *b, void *handle)
+plugin_can_extents (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_extents)
-    return normalize_bool (p->plugin.can_extents (handle));
+    return normalize_bool (p->plugin.can_extents (c->handle));
   else
     return p->plugin.extents != NULL;
 }
 
 static int
-plugin_can_fua (struct backend *b, void *handle)
+plugin_can_fua (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
   /* The plugin must use API version 2 and have .can_fua return
      NBDKIT_FUA_NATIVE before we will pass the FUA flag on. */
   if (p->plugin.can_fua) {
-    r = p->plugin.can_fua (handle);
+    r = p->plugin.can_fua (c->handle);
     if (r > NBDKIT_FUA_EMULATE && p->plugin._api_version == 1)
       r = NBDKIT_FUA_EMULATE;
     return r;
@@ -517,23 +603,25 @@ plugin_can_fua (struct backend *b, void *handle)
 }
 
 static int
-plugin_can_multi_conn (struct backend *b, void *handle)
+plugin_can_multi_conn (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_multi_conn)
-    return normalize_bool (p->plugin.can_multi_conn (handle));
+    return normalize_bool (p->plugin.can_multi_conn (c->handle));
   else
     return 0; /* assume false */
 }
 
 static int
-plugin_can_cache (struct backend *b, void *handle)
+plugin_can_cache (struct context *c)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   if (p->plugin.can_cache)
-    return p->plugin.can_cache (handle);
+    return p->plugin.can_cache (c->handle);
   if (p->plugin.cache)
     return NBDKIT_CACHE_NATIVE;
   return NBDKIT_CACHE_NONE;
@@ -542,7 +630,7 @@ plugin_can_cache (struct backend *b, void *handle)
 /* Plugins and filters can call this to set the true errno, in cases
  * where !errno_is_preserved.
  */
-void
+NBDKIT_DLL_PUBLIC void
 nbdkit_set_error (int err)
 {
   threadlocal_set_error (err);
@@ -561,35 +649,37 @@ get_error (struct backend_plugin *p)
 }
 
 static int
-plugin_pread (struct backend *b, void *handle,
+plugin_pread (struct context *c,
               void *buf, uint32_t count, uint64_t offset, uint32_t flags,
               int *err)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
   assert (p->plugin.pread || p->plugin._pread_v1);
 
   if (p->plugin.pread)
-    r = p->plugin.pread (handle, buf, count, offset, 0);
+    r = p->plugin.pread (c->handle, buf, count, offset, 0);
   else
-    r = p->plugin._pread_v1 (handle, buf, count, offset);
+    r = p->plugin._pread_v1 (c->handle, buf, count, offset);
   if (r == -1)
     *err = get_error (p);
   return r;
 }
 
 static int
-plugin_flush (struct backend *b, void *handle,
+plugin_flush (struct context *c,
               uint32_t flags, int *err)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
   if (p->plugin.flush)
-    r = p->plugin.flush (handle, 0);
+    r = p->plugin.flush (c->handle, 0);
   else if (p->plugin._flush_v1)
-    r = p->plugin._flush_v1 (handle);
+    r = p->plugin._flush_v1 (c->handle);
   else {
     *err = EINVAL;
     return -1;
@@ -600,66 +690,69 @@ plugin_flush (struct backend *b, void *handle,
 }
 
 static int
-plugin_pwrite (struct backend *b, void *handle,
+plugin_pwrite (struct context *c,
                const void *buf, uint32_t count, uint64_t offset, uint32_t flags,
                int *err)
 {
+  struct backend *b = c->b;
   int r;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   bool fua = flags & NBDKIT_FLAG_FUA;
   bool need_flush = false;
 
-  if (fua && backend_can_fua (b) != NBDKIT_FUA_NATIVE) {
+  if (fua && backend_can_fua (c) != NBDKIT_FUA_NATIVE) {
     flags &= ~NBDKIT_FLAG_FUA;
     need_flush = true;
   }
   if (p->plugin.pwrite)
-    r = p->plugin.pwrite (handle, buf, count, offset, flags);
+    r = p->plugin.pwrite (c->handle, buf, count, offset, flags);
   else if (p->plugin._pwrite_v1)
-    r = p->plugin._pwrite_v1 (handle, buf, count, offset);
+    r = p->plugin._pwrite_v1 (c->handle, buf, count, offset);
   else {
     *err = EROFS;
     return -1;
   }
   if (r != -1 && need_flush)
-    r = plugin_flush (b, handle, 0, err);
+    r = plugin_flush (c, 0, err);
   if (r == -1 && !*err)
     *err = get_error (p);
   return r;
 }
 
 static int
-plugin_trim (struct backend *b, void *handle,
+plugin_trim (struct context *c,
              uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
+  struct backend *b = c->b;
   int r;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   bool fua = flags & NBDKIT_FLAG_FUA;
   bool need_flush = false;
 
-  if (fua && backend_can_fua (b) != NBDKIT_FUA_NATIVE) {
+  if (fua && backend_can_fua (c) != NBDKIT_FUA_NATIVE) {
     flags &= ~NBDKIT_FLAG_FUA;
     need_flush = true;
   }
   if (p->plugin.trim)
-    r = p->plugin.trim (handle, count, offset, flags);
+    r = p->plugin.trim (c->handle, count, offset, flags);
   else if (p->plugin._trim_v1)
-    r = p->plugin._trim_v1 (handle, count, offset);
+    r = p->plugin._trim_v1 (c->handle, count, offset);
   else {
     *err = EINVAL;
     return -1;
   }
   if (r != -1 && need_flush)
-    r = plugin_flush (b, handle, 0, err);
+    r = plugin_flush (c, 0, err);
   if (r == -1 && !*err)
     *err = get_error (p);
   return r;
 }
 
 static int
-plugin_zero (struct backend *b, void *handle,
+plugin_zero (struct context *c,
              uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r = -1;
   bool may_trim = flags & NBDKIT_FLAG_MAY_TRIM;
@@ -668,23 +761,23 @@ plugin_zero (struct backend *b, void *handle,
   bool emulate = false;
   bool need_flush = false;
 
-  if (fua && backend_can_fua (b) != NBDKIT_FUA_NATIVE) {
+  if (fua && backend_can_fua (c) != NBDKIT_FUA_NATIVE) {
     flags &= ~NBDKIT_FLAG_FUA;
     need_flush = true;
   }
   if (!count)
     return 0;
 
-  if (backend_can_zero (b) == NBDKIT_ZERO_NATIVE) {
+  if (backend_can_zero (c) == NBDKIT_ZERO_NATIVE) {
     errno = 0;
     if (p->plugin.zero)
-      r = p->plugin.zero (handle, count, offset, flags);
+      r = p->plugin.zero (c->handle, count, offset, flags);
     else if (p->plugin._zero_v1) {
       if (fast_zero) {
         *err = EOPNOTSUPP;
         return -1;
       }
-      r = p->plugin._zero_v1 (handle, count, offset, may_trim);
+      r = p->plugin._zero_v1 (c->handle, count, offset, may_trim);
     }
     else
       emulate = true;
@@ -711,7 +804,7 @@ plugin_zero (struct backend *b, void *handle,
     static /* const */ char buf[MAX_REQUEST_SIZE];
     uint32_t limit = MIN (count, sizeof buf);
 
-    r = plugin_pwrite (b, handle, buf, limit, offset, flags, err);
+    r = plugin_pwrite (c, buf, limit, offset, flags, err);
     if (r == -1)
       break;
     count -= limit;
@@ -719,17 +812,18 @@ plugin_zero (struct backend *b, void *handle,
 
  done:
   if (r != -1 && need_flush)
-    r = plugin_flush (b, handle, 0, err);
+    r = plugin_flush (c, 0, err);
   if (r == -1 && !*err)
     *err = get_error (p);
   return r;
 }
 
 static int
-plugin_extents (struct backend *b, void *handle,
+plugin_extents (struct context *c,
                 uint32_t count, uint64_t offset, uint32_t flags,
                 struct nbdkit_extents *extents, int *err)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
@@ -738,7 +832,7 @@ plugin_extents (struct backend *b, void *handle,
     return -1;
   }
 
-  r = p->plugin.extents (handle, count, offset, flags, extents);
+  r = p->plugin.extents (c->handle, count, offset, flags, extents);
   if (r >= 0 && nbdkit_extents_count (extents) < 1) {
     nbdkit_error ("extents: plugin must return at least one extent");
     nbdkit_set_error (EINVAL);
@@ -750,10 +844,11 @@ plugin_extents (struct backend *b, void *handle,
 }
 
 static int
-plugin_cache (struct backend *b, void *handle,
+plugin_cache (struct context *c,
               uint32_t count, uint64_t offset, uint32_t flags,
               int *err)
 {
+  struct backend *b = c->b;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   int r;
 
@@ -762,7 +857,7 @@ plugin_cache (struct backend *b, void *handle,
   if (!p->plugin.cache)
     return 0;
 
-  r = p->plugin.cache (handle, count, offset, flags);
+  r = p->plugin.cache (c->handle, count, offset, flags);
   if (r == -1)
     *err = get_error (p);
   return r;
@@ -780,6 +875,7 @@ static struct backend plugin_functions = {
   .magic_config_key = plugin_magic_config_key,
   .get_ready = plugin_get_ready,
   .after_fork = plugin_after_fork,
+  .cleanup = plugin_cleanup,
   .preconnect = plugin_preconnect,
   .list_exports = plugin_list_exports,
   .default_export = plugin_default_export,
@@ -789,6 +885,7 @@ static struct backend plugin_functions = {
   .close = plugin_close,
   .export_description = plugin_export_description,
   .get_size = plugin_get_size,
+  .block_size = plugin_block_size,
   .can_write = plugin_can_write,
   .can_flush = plugin_can_flush,
   .is_rotational = plugin_is_rotational,
@@ -811,7 +908,7 @@ static struct backend plugin_functions = {
 /* Register and load a plugin. */
 struct backend *
 plugin_register (size_t index, const char *filename,
-                 void *dl, struct nbdkit_plugin *(*plugin_init) (void))
+                 void *dl, plugin_init_function plugin_init)
 {
   struct backend_plugin *p;
   const struct nbdkit_plugin *plugin;
